def list_tasks():
    return '''classify: Given text, classify the text.
    parse: Given a prompt and a context, parse the answer from the context.
    generate: Given a prompt, generate an answer.'''